"""Module containing all of the BREWMASTER error definitions"""


# Models
class BrewmasterModelError(Exception):
    """Wrapper Error for All BrewmasterModelErrors"""
    pass


class BrewmasterModelValidationError(BrewmasterModelError):
    """Error to indicate an invalid Brewmaster Model"""
    pass


class RequestStatusTransitionError(BrewmasterModelValidationError):
    """Error to indicate an updated status was not a valid transition"""
    pass


# Plugins
class PluginError(Exception):
    """Generic error class"""
    pass


class PluginValidationError(PluginError):
    """Plugin could not be validated successfully"""
    pass


class PluginParamError(PluginError):
    """Error used when plugins have illegal parameters"""
    pass


# Requests
class AckAndContinueException(Exception):
    pass


class NoAckAndDieException(Exception):
    pass


class AckAndDieException(Exception):
    pass


class DiscardMessageException(Exception):
    """Raising an instance will result in a message not being requeued"""
    pass


class RepublishRequestException(Exception):
    """Republish to the end of the message queue

    :param request: The Request to republish
    :param headers: A dictionary of headers to be used by
        `brewtils.request_consumer.RequestConsumer`
    :type request: :py:class:`brewtils.models.Request`
    """
    def __init__(self, request, headers):
        self.request = request
        self.headers = headers


class RequestProcessingError(AckAndContinueException):
    pass


# Rest / Client errors
class BrewmasterRestError(Exception):
    """Wrapper Error to Wrap more specific BREWMASTER Rest Errors"""
    pass


class BrewmasterConnectionError(BrewmasterRestError):
    """Error indicating a connection error while performing a request"""
    pass


class BrewmasterTimeoutError(BrewmasterRestError):
    """Error Indicating a Timeout was reached while performing a request"""
    pass


class BrewmasterFetchError(BrewmasterRestError):
    """Error Indicating a server Error occurred performing a GET"""
    pass


class BrewmasterValidationError(BrewmasterRestError):
    """Error Indicating a client (400) Error occurred performing a POST/PUT"""
    pass


class BrewmasterSaveError(BrewmasterRestError):
    """Error Indicating a server Error occurred performing a POST/PUT"""
    pass


class BrewmasterDeleteError(BrewmasterRestError):
    """Error Indicating a server Error occurred performing a DELETE"""
    pass


class BGConflictError(BrewmasterRestError):
    """Error indicating a 409 was raised on the server"""
    pass


class BGNotFoundError(BrewmasterRestError):
    """Error Indicating a 404 was raised on the server"""
    pass



import logging
import warnings

from brewtils.models import System, Instance, Command, Parameter, Request, PatchOperation, \
    Choices, LoggingConfig, Event, Queue
from brewtils.schemas import SystemSchema, InstanceSchema, CommandSchema, ParameterSchema, \
    RequestSchema, PatchSchema, LoggingConfigSchema, EventSchema, QueueSchema


class SchemaParser(object):
    """Serialize and deserialize Brewtils models"""

    _models = {
        'SystemSchema': System,
        'InstanceSchema': Instance,
        'CommandSchema': Command,
        'ParameterSchema': Parameter,
        'RequestSchema': Request,
        'PatchSchema': PatchOperation,
        'ChoicesSchema': Choices,
        'LoggingConfigSchema': LoggingConfig,
        'EventSchema': Event,
        'QueueSchema': Queue
    }

    logger = logging.getLogger(__name__)

    # Deserialization methods
    @classmethod
    def parse_system(cls, system, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a system model object

        :param system: The raw input
        :param from_string: True if 'system' is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A System object
        """
        return cls._do_parse(system, SystemSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_instance(cls, instance, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to an instance model object

        :param instance: The raw input
        :param from_string: True if 'instance' is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: An Instance object
        """
        return cls._do_parse(instance, InstanceSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_command(cls, command, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a command model object

        :param command: The raw input
        :param from_string: True if 'command' is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A Command object
        """
        return cls._do_parse(command, CommandSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_parameter(cls, parameter, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a parameter model object

        :param parameter: The raw input
        :param from_string: True if 'parameter' is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: An Parameter object
        """
        return cls._do_parse(parameter, ParameterSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_request(cls, request, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a request model object

        :param request: The raw input
        :param from_string: True if 'request' is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A Request object
        """
        return cls._do_parse(request, RequestSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_patch(cls, patch, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a patch model object

        Note: for our patches, many is _always_ set to True. We will always return a list
        from this method.

        :param patch: The raw input
        :param from_string: True if 'patch' is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A PatchOperation object
        """
        if not kwargs.pop('many', True):
            cls.logger.warning("A patch object should always be wrapped as a list of objects. "
                               "Thus, parsing will always return a list. You specified many as "
                               "False, this is being ignored and a list "
                               "will be returned anyway.")
        return cls._do_parse(patch, PatchSchema(many=True, **kwargs), from_string=from_string)

    @classmethod
    def parse_logging_config(cls, logging_config, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a logging config model object

        Note: for our logging_config, many is _always_ set to False. We will always return a dict
        from this method.

        :param logging_config: The raw input
        :param from_string: True if 'logging_config' is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A LoggingConfig object
        """
        if kwargs.pop('many', False):
            cls.logger.warning("A logging config object should never be wrapped as a list of "
                               "objects. Thus, parsing will always return a dict. You specified "
                               "many as True, this is being ignored and a dict will be returned "
                               "anyway.")
        return cls._do_parse(logging_config,
                             LoggingConfigSchema(many=False, **kwargs),
                             from_string=from_string)

    @classmethod
    def parse_event(cls, event, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to an event model object

        :param event: The raw input
        :param from_string: True if 'event' is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: An Event object
        """
        return cls._do_parse(event, EventSchema(**kwargs), from_string=from_string)

    @classmethod
    def parse_queue(cls, queue, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a queue model object

        :param queue: The raw input
        :param from_string: True if 'event' is a JSON string, False if a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: A Queue object
        """
        return cls._do_parse(queue, QueueSchema(**kwargs), from_string=from_string)

    @classmethod
    def _do_parse(cls, data, schema, from_string=False):
        schema.context['models'] = cls._models
        return schema.loads(data).data if from_string else schema.load(data).data

    # Serialization methods
    @classmethod
    def serialize_system(cls, system, to_string=True, include_commands=True, **kwargs):
        """Convert a system model into serialized form

        :param system: The system object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param include_commands: True if the system's command list should be included
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of system
        """
        if not include_commands:
            if 'exclude' in kwargs:
                kwargs['exclude'] += ('commands', )
            else:
                kwargs['exclude'] = ('commands', )

        return cls._do_serialize(SystemSchema(**kwargs), system, to_string)

    @classmethod
    def serialize_instance(cls, instance, to_string=True, **kwargs):
        """Convert an instance model into serialized form

        :param instance: The instance object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of instance
        """
        return cls._do_serialize(InstanceSchema(**kwargs), instance, to_string)

    @classmethod
    def serialize_command(cls, command, to_string=True, **kwargs):
        """Convert a command model into serialized form

        :param command: The command object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of command
        """
        return cls._do_serialize(CommandSchema(**kwargs), command, to_string)

    @classmethod
    def serialize_parameter(cls, parameter, to_string=True, **kwargs):
        """Convert a parameter model into serialized form

        :param parameter: The parameter object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of parameter
        """
        return cls._do_serialize(ParameterSchema(**kwargs), parameter, to_string)

    @classmethod
    def serialize_request(cls, request, to_string=True, **kwargs):
        """Convert a request model into serialized form

        :param request: The request object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of request
        """
        return cls._do_serialize(RequestSchema(**kwargs), request, to_string)

    @classmethod
    def serialize_patch(cls, patch, to_string=True, **kwargs):
        """Convert a patch model into serialized form

        :param patch: The patch object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of patch
        """
        return cls._do_serialize(PatchSchema(**kwargs), patch, to_string)

    @classmethod
    def serialize_logging_config(cls, logging_config, to_string=True, **kwargs):
        """Convert a logging config model into serialize form

        :param logging_config: The logging config object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of logging config
        """
        return cls._do_serialize(LoggingConfigSchema(**kwargs), logging_config, to_string)

    @classmethod
    def serialize_event(cls, event, to_string=True, **kwargs):
        """Convert a logging config model into serialized form

        :param event: The event object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of event
        """
        return cls._do_serialize(EventSchema(**kwargs), event, to_string)

    @classmethod
    def serialize_queue(cls, queue, to_string=True, **kwargs):
        """Convert a queue model into serialized form

        :param queue: The queue object(s) to be serialized
        :param to_string: True to generate a JSON-formatted string, False to generate a dictionary
        :param kwargs: Additional parameters to be passed to the Schema (e.g. many=True)
        :return: Serialized representation of queue
        """
        return cls._do_serialize(QueueSchema(**kwargs), queue, to_string)

    @staticmethod
    def _do_serialize(schema, data, to_string):
        return schema.dumps(data).data if to_string else schema.dump(data).data


class BrewmasterSchemaParser(SchemaParser):
    def __init__(self):
        warnings.warn("Reference made to 'BrewmasterSchemaParser'. This name will be removed in "
                      "version 3.0, please use 'SchemaParser' instead.",
                      DeprecationWarning, stacklevel=2)
        super(BrewmasterSchemaParser, self).__init__()



import os

import six

from brewtils.rest import normalize_url_prefix
from ._version import __version__ as generated_version

__version__ = generated_version


def get_easy_client(**kwargs):
    """Initialize an EasyClient using Environment variables as default values

    :param kwargs: Options for configuring the EasyClient
    :return: An EasyClient
    """
    from brewtils.rest.easy_client import EasyClient

    parser = kwargs.pop('parser', None)
    logger = kwargs.pop('logger', None)

    return EasyClient(logger=logger, parser=parser, **get_bg_connection_parameters(**kwargs))


def get_bool_from_kwargs_and_env(key, env_name, **kwargs):
    """Gets a boolean value defaults to True"""
    value = kwargs.get(key, None)
    if value is None:
        return os.environ.get(env_name, 'true').lower() != 'false'
    elif isinstance(value, six.string_types):
        return value.lower() != 'false'
    else:
        return bool(value)


def get_from_kwargs_or_env(key, env_names, default, **kwargs):
    """Get a value from the kwargs provided or environment

    :param key: Key to search in the keyword args
    :param env_names: Environment names to search
    :param default: The default if it is not found elsewhere
    :param kwargs: Keyword Arguments
    :return:
    """
    if kwargs.get(key, None) is not None:
        return kwargs[key]

    for name in env_names:
        if name in os.environ:
            return os.environ[name]

    return default


def get_bg_connection_parameters(**kwargs):
    """Parse the keyword arguments, search in the arguments, and environment for the values

    :param kwargs:
    :return:
    """
    from brewtils.rest.client import RestClient
    from brewtils.errors import BrewmasterValidationError

    host = kwargs.pop('host', None) or os.environ.get('BG_WEB_HOST')
    if not host:
        raise BrewmasterValidationError('Unable to create a plugin without a beer-garden host. '
                                        'Please specify one with bg_host=<host> or by setting the '
                                        'BG_WEB_HOST environment variable.')

    port = get_from_kwargs_or_env('port', ['BG_WEB_PORT'], '2337', **kwargs)

    url_prefix = get_from_kwargs_or_env('url_prefix', ['BG_URL_PREFIX'], None, **kwargs)
    url_prefix = normalize_url_prefix(url_prefix)

    ssl_enabled = get_bool_from_kwargs_and_env('ssl_enabled', 'BG_SSL_ENABLED', **kwargs)
    ca_verify = get_bool_from_kwargs_and_env('ca_verify', 'BG_CA_VERIFY', **kwargs)

    api_version = kwargs.pop('api_version', RestClient.LATEST_VERSION)
    ca_cert = get_from_kwargs_or_env('ca_cert', ['BG_CA_CERT', 'BG_SSL_CA_CERT'], None, **kwargs)
    client_cert = get_from_kwargs_or_env('client_cert', ['BG_CLIENT_CERT', 'BG_SSL_CLIENT_CERT'],
                                         None, **kwargs)

    return {
        'host': host,
        'port': port,
        'ssl_enabled': ssl_enabled,
        'api_version': api_version,
        'ca_cert': ca_cert,
        'client_cert': client_cert,
        'url_prefix': url_prefix,
        'ca_verify': ca_verify
    }



import logging
import threading
from functools import partial

import pika
from pika.exceptions import AMQPConnectionError

from brewtils.errors import DiscardMessageException, RepublishRequestException
from brewtils.schema_parser import SchemaParser


class RequestConsumer(threading.Thread):
    """Consumer that will handle unexpected interactions with RabbitMQ.

    If RabbitMQ closes the connection, it will reopen it. You should look at the output, as there
    are limited reasons why the connection may be closed, which usually are tied to permission
    related issues or socket timeouts.

    If the channel is closed, it will indicate a problem with one of the commands that were issued
    and that should surface in the output as well.

    :param str amqp_url: The AMQP url to connection with
    :param str queue_name: The name of the queue to connect to
    :param func on_message_callback: The function called to invoke message processing.
        Must return a Future.
    :param event panic_event: An event to be set in the event of a catastrophic failure
    :type event: :py:class:`threading.Event`
    :param logger: A configured logger
    :type logger: :py:class:`logging.Logger`
    :param str thread_name: The name to use for this thread
    :param int max_connect_retries: Number of connection retry attempts before failure.
        Default -1 (retry forever).
    :param int max_connect_backoff: Maximum amount of time to wait between connection retry
        attempts. Default 30.
    :param int max_concurrent: Maximum number of requests to process concurrently
    """

    def __init__(self, amqp_url, queue_name, on_message_callback, panic_event,
                 logger=None, thread_name=None, **kwargs):

        self._connection = None
        self._channel = None
        self._consumer_tag = None

        self._url = amqp_url
        self._queue_name = queue_name
        self._on_message_callback = on_message_callback
        self._panic_event = panic_event
        self._max_connect_retries = kwargs.pop("max_connect_retries", -1)
        self._max_connect_backoff = kwargs.pop("max_connect_backoff", 30)
        self._max_concurrent = kwargs.pop("max_concurrent", 1)
        self.logger = logger or logging.getLogger(__name__)
        self.shutdown_event = threading.Event()

        super(RequestConsumer, self).__init__(name=thread_name)

    def run(self):
        """Run the example consumer.

        Creates a connection to the queueing service, then starts the IOLoop. The IOLoop will block
        and allow the SelectConnection to operate.

        :return:
        """
        self._connection = self.open_connection()

        # It is possible to return from open_connection without acquiring a connection.
        # This usually happens if no max_connect_retries was set and we are constantly trying to
        # connect to a queue that does not exist. For those cases, there is no reason to start
        # an ioloop.
        if self._connection:
            self._connection.ioloop.start()

    def stop(self):
        """Cleanly shutdown the connection.

        Assumes the stop_consuming method has already been called. When the queueing service
        acknowledges the closure, the connection is closed which will end the RequestConsumer.

        :return:
        """
        self.logger.debug('Stopping request consumer')
        self.shutdown_event.set()
        self.close_channel()

    def on_message(self, channel, basic_deliver, properties, body):
        """Invoked when a message is delivered from the queueing service.

        Invoked by pika when a message is delivered from RabbitMQ. The channel is passed for your
        convenience. The basic_deliver object that is passed in carries the exchange, routing key,
        delivery tag and a redelivered flag for the message. the properties passed in is an
        instance of BasicProperties with the message properties and the body is the message that
        was sent.

        :param pika.channel.Channel channel: The channel object
        :param pika.Spec.Basic.Deliver basic_deliver: basic_deliver method
        :param pika.Spec.BasicProperties properties: properties
        :param str|unicode body: The message body
        """
        self.logger.debug("Received message #%s from %s on channel %s: %s",
                          basic_deliver.delivery_tag, properties.app_id,
                          channel.channel_number, body)

        try:
            future = self._on_message_callback(body, properties.headers)
            future.add_done_callback(partial(self.on_message_callback_complete, basic_deliver))
        except DiscardMessageException:
            self.logger.debug('Nacking message %s, not attempting to requeue',
                              basic_deliver.delivery_tag)
            self._channel.basic_nack(basic_deliver.delivery_tag, requeue=False)
        except Exception as ex:
            self.logger.exception('Exception while trying to schedule message %s, about to nack '
                                  'and requeue: %s',
                                  basic_deliver.delivery_tag, ex)
            self._channel.basic_nack(basic_deliver.delivery_tag, requeue=True)

    def on_message_callback_complete(self, basic_deliver, future):
        """Invoked when the future returned by _on_message_callback completes.

        :param pika.Spec.Basic.Deliver basic_deliver: basic_deliver method
        :param concurrent.futures.Future future: Completed future
        :return: None
        """
        if not future.exception():
            try:
                self.logger.debug('Acking message %s', basic_deliver.delivery_tag)
                self._channel.basic_ack(basic_deliver.delivery_tag)
            except Exception as ex:
                self.logger.exception('Error acking message %s, about to shut down: %s',
                                      basic_deliver.delivery_tag, ex)
                self._panic_event.set()
        else:
            future_ex = future.exception()

            if isinstance(future_ex, RepublishRequestException):
                try:
                    with pika.BlockingConnection(pika.URLParameters(self._url)) as conn:
                        headers = future_ex.headers
                        headers.update({'request_id': future_ex.request.id})
                        props = pika.BasicProperties(app_id='beer-garden',
                                                     content_type='text/plain',
                                                     headers=headers, priority=1)
                        conn.channel().basic_publish(exchange=basic_deliver.exchange,
                                                     properties=props,
                                                     routing_key=basic_deliver.routing_key,
                                                     body=SchemaParser.serialize_request(
                                                         future_ex.request)
                                                     )

                    self._channel.basic_ack(basic_deliver.delivery_tag)
                except Exception as ex:
                    self.logger.exception('Error republishing message %s, about to shut down: %s',
                                          basic_deliver.delivery_tag, ex)
                    self._panic_event.set()
            elif isinstance(future_ex, DiscardMessageException):
                self.logger.info('Nacking message %s, not attempting to requeue',
                                 basic_deliver.delivery_tag)
                self._channel.basic_nack(basic_deliver.delivery_tag, requeue=False)
            else:
                # If request processing throws anything else we are in a seriously bad state
                self.logger.exception('Unexpected exception during request %s processing, '
                                      'about to shut down: %s',
                                      basic_deliver.delivery_tag, future_ex, exc_info=False)
                self._panic_event.set()

    def open_connection(self):
        """Opens a connection to the queueing service.

        This method connects to RabbitMQ, returning the connection handle. When the connection
        is established, the on_connection_open method will be invoked by pika.

        :rtype: pika.SelectConnection
        """
        self.logger.debug('Connecting to %s' % self._url)
        time_to_wait = 0.1
        retries = 0
        while not self.shutdown_event.is_set():
            try:
                return pika.SelectConnection(pika.URLParameters(self._url), self.on_connection_open,
                                             stop_ioloop_on_close=False)
            except AMQPConnectionError as ex:
                if 0 <= self._max_connect_retries <= retries:
                    raise ex
                self.logger.warning("Error attempting to connect to %s" % self._url)
                self.logger.warning("Waiting %s seconds and attempting again" % time_to_wait)
                self.shutdown_event.wait(time_to_wait)
                time_to_wait = min(time_to_wait * 2, self._max_connect_backoff)
                retries += 1

    def on_connection_open(self, unused_connection):
        """Invoked when the connection has been established.

        This method is called by pika once the connection to RabbitMQ has been established. It
        passes the handle to the connection object in case we need it, but in this case, we'll
        just mark it unused.

        :type unused_connection: pika.SelectConnection
        """
        self.logger.debug("Connection opened: %s", unused_connection)
        self._connection.add_on_close_callback(self.on_connection_closed)
        self.open_channel()

    def close_connection(self):
        """This method closes the connection to RabbitMQ."""
        self.logger.debug("Closing connection")
        self._connection.close()

    def on_connection_closed(self, connection, reply_code, reply_text):
        """Invoked when the connection is closed.

        This method is invoked by pika when the connection to RabbitMQ is closed unexpectedly.
        Since it is unexpected, we will reconnect to RabbitMQ if it disconnects.

        :param pika.connection.Connection connection: the closed connection object
        :param int reply_code: The server provided reply_code if given
        :param basestring reply_text: The server provided reply_text if given
        """
        self.logger.debug('Connection "%s" closed: (%s) %s' % (connection, reply_code, reply_text))
        self._channel = None

        # A 320 is the server forcing the connection to close
        if reply_code == 320:
            self.shutdown_event.set()

        if self.shutdown_event.is_set():
            self._connection.ioloop.stop()
        else:
            self.logger.warning('Connection unexpectedly closed: (%s) %s' %
                                (reply_code, reply_text))
            self.logger.warning('Attempting to reopen connection in 5 seconds')
            self._connection.add_timeout(5, self.reconnect)

    def reconnect(self):
        """Will be invoked by the IOLoop timer if the connection is closed."""

        # This is the old connection IOLoop instance, stop its ioloop
        self._connection.ioloop.stop()

        if not self.shutdown_event.is_set():
            # Creates a new connection
            self._connection = self.open_connection()

            # There is now a new connection, needs a new ioloop to run
            if self._connection:
                self._connection.ioloop.start()

    def open_channel(self):
        """Opens a channel on the queueing service.

        Open a new channel with RabbitMQ by issuing the Channel.Open RPC command.
        When RabbitMQ responds that the
        channel is open, the on_channel_open callback will be invoked by pika.
        """
        self.logger.debug('Opening a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.

        The exchange / queue binding should have already been set up so
        just start consuming.

        :param pika.channel.Channel channel: The channel object
        """
        self.logger.debug('Channel opened: %s', channel)
        self._channel = channel
        self._channel.add_on_close_callback(self.on_channel_closed)
        self.start_consuming()

    def close_channel(self):
        """Call to close the channel cleanly by issuing the Channel.Close RPC command."""
        self.logger.debug('Closing the channel')
        self._channel.close()

    def on_channel_closed(self, channel, reply_code, reply_text):
        """Invoekd when the queueing service unexpectedly closes the channle.

        Invoked by pika when RabbitMQ unexpectedly closes the channel. Channels are usually closed
        if you attempt to do something that violates the protocol, such as re-declare an exchange
        or queue with different parameters. In this case, we'll close the connection to shutdown
        the object.

        :param pika.channel.Channel channel: The closed channel
        :param int reply_code: The numeric reason the channel was closed
        :param str reply_text: The text reason the channel was closed
        """
        self.logger.debug('Channel %i was closed: (%s) %s' % (channel, reply_code, reply_text))
        self._connection.close()

    def start_consuming(self):
        """Begin consuming messages from the queueing service.

        This method sets up the consumer by first calling add_on_cancel_callback so that the
        object is notified if RabbitMQ cancels the consumer. It then issues the Basic.Consume RPC
        command which returns the consumer tag that is used to uniquely identify the consumer with
        RabbitMQ. We keep the value to use it when we want to cancel consuming. The on_message
        method is passed in as a callback pika will invoke when a message is fully received.
        """
        self.logger.debug('Issuing consumer related RPC commands')

        # Prefetch of 1 to prevent RabbitMQ from sending us multiple messages at once
        self._channel.basic_qos(prefetch_count=self._max_concurrent)
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)
        self._consumer_tag = self._channel.basic_consume(self.on_message, queue=self._queue_name)

    def stop_consuming(self):
        """Stop consuming by sending the Basic.Cancel RPC command."""
        self.logger.debug("Stopping consuming on channel %s", self._channel)
        if self._channel:
            self.logger.debug('Sending a Basic.Cancel RPC command to RabbitMQ')
            self._channel.basic_cancel(self.on_cancelok, self._consumer_tag)

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        self.logger.debug('Consumer was cancelled remotely, shutting down: %r' % method_frame)
        if self._channel:
            self.close_channel()

    def on_cancelok(self, unused_frame):
        """Invoked when the queueing service acknowledges cancellation.

        This method is invoked by pika when RabbitMQ acknowledges the cancellation of a consumer.
        At this point we will close the channel. This will invoke the on_channel_closed method
        once the channel has been closed, which will in-turn close the connection.

        :param pika.frame.Method unused_frame: The Basic.CancelOK frame
        """
        self.logger.debug(unused_frame)
        self.logger.debug('RabbitMQ acknowledged the cancellation of the consumer')



from enum import Enum

import six
from brewtils.errors import RequestStatusTransitionError


class Events(Enum):
    BREWVIEW_STARTED = 1
    BREWVIEW_STOPPED = 2
    BARTENDER_STARTED = 3
    BARTENDER_STOPPED = 4
    REQUEST_CREATED = 5
    REQUEST_STARTED = 6
    REQUEST_COMPLETED = 7
    INSTANCE_INITIALIZED = 8
    INSTANCE_STARTED = 9
    INSTANCE_STOPPED = 10
    SYSTEM_CREATED = 11
    SYSTEM_UPDATED = 12
    SYSTEM_REMOVED = 13
    QUEUE_CLEARED = 14
    ALL_QUEUES_CLEARED = 15


class Command(object):

    schema = 'CommandSchema'

    COMMAND_TYPES = ('ACTION', 'INFO', 'EPHEMERAL')
    OUTPUT_TYPES = ('STRING', 'JSON', 'XML', 'HTML')

    def __init__(self, name, description=None, id=None, parameters=None, command_type=None,
                 output_type=None, schema=None, form=None, template=None, icon_name=None,
                 system=None):
        self.name = name
        self.description = description
        self.id = id
        self.parameters = parameters or []
        self.command_type = command_type
        self.output_type = output_type
        self.schema = schema
        self.form = form
        self.template = template
        self.icon_name = icon_name
        self.system = system

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<Command: %s>' % self.name

    def parameter_keys(self):
        """Convenience Method for returning all the keys of this command's parameters.

        :return list_of_parameters:
        """
        return [p.key for p in self.parameters]

    def get_parameter_by_key(self, key):
        """Given a Key, it will return the parameter (or None) with that key

        :param key:
        :return parameter:
        """
        for parameter in self.parameters:
            if parameter.key == key:
                return parameter

        return None

    def has_different_parameters(self, parameters):
        """Given a set of parameters, determines if the parameters provided differ from the
        parameters already defined on this command.

        :param parameters:
        :return boolean:
        """
        if len(parameters) != len(self.parameters):
            return True

        for parameter in parameters:
            if parameter.key not in self.parameter_keys():
                return True

            current_param = self.get_parameter_by_key(parameter.key)
            if current_param.is_different(parameter):
                return True

        return False


class Instance(object):

    schema = 'InstanceSchema'

    INSTANCE_STATUSES = {'INITIALIZING', 'RUNNING', 'PAUSED', 'STOPPED', 'DEAD', 'UNRESPONSIVE',
                         'STARTING', 'STOPPING', 'UNKNOWN'}

    def __init__(self, name=None, description=None, id=None, status=None, status_info=None,
                 queue_type=None, queue_info=None, icon_name=None, metadata=None):

        self.name = name
        self.description = description
        self.id = id
        self.status = status.upper() if status else None
        self.status_info = status_info or {}
        self.queue_type = queue_type
        self.queue_info = queue_info or {}
        self.icon_name = icon_name
        self.metadata = metadata or {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<Instance: name=%s, status=%s>' % (self.name, self.status)


class Choices(object):

    schema = 'ChoicesSchema'

    TYPES = ('static', 'url', 'command')
    DISPLAYS = ('select', 'typeahead')

    def __init__(self, type=None, display=None, value=None, strict=None, details=None):
        self.type = type
        self.strict = strict
        self.value = value
        self.display = display
        self.details = details or {}

    def __str__(self):
        return self.value.__str__()

    def __repr__(self):
        return '<Choices: type=%s, display=%s, value=%s>' % (self.type, self.display, self.value)


class Parameter(object):

    schema = 'ParameterSchema'

    TYPES = ("String", "Integer", "Float", "Boolean", "Any", "Dictionary", "Date", "DateTime")
    FORM_INPUT_TYPES = ("textarea",)

    def __init__(self, key, type=None, multi=None, display_name=None, optional=None, default=None,
                 description=None, choices=None, parameters=None, nullable=None, maximum=None,
                 minimum=None, regex=None, form_input_type=None):

        self.key = key
        self.type = type
        self.multi = multi
        self.display_name = display_name
        self.optional = optional
        self.default = default
        self.description = description
        self.choices = choices
        self.parameters = parameters or []
        self.nullable = nullable
        self.maximum = maximum
        self.minimum = minimum
        self.regex = regex
        self.form_input_type = form_input_type

    def __str__(self):
        return self.key

    def __repr__(self):
        return '<Parameter: key=%s, type=%s, description=%s>' % (self.key,
                                                                 self.type,
                                                                 self.description)

    def is_different(self, other):
        if not type(other) is type(self):
            return True

        fields_to_compare = ['key', 'type', 'multi', 'optional', 'default', 'nullable', 'maximum',
                             'minimum', 'regex']
        for field in fields_to_compare:
            if getattr(self, field) != getattr(other, field):
                return True

        if len(self.parameters) != len(other.parameters):
            return True

        parameter_keys = [p.key for p in self.parameters]
        for parameter in other.parameters:
            if parameter.key not in parameter_keys:
                return True

            current_param = list(filter((lambda p: p.key == parameter.key), self.parameters))[0]
            if current_param.is_different(parameter):
                return True

        return False


class Request(object):

    schema = 'RequestSchema'

    STATUS_LIST = ('CREATED', 'RECEIVED', 'IN_PROGRESS', 'CANCELED', 'SUCCESS', 'ERROR')
    COMPLETED_STATUSES = ('CANCELED', 'SUCCESS', 'ERROR')
    COMMAND_TYPES = ('ACTION', 'INFO', 'EPHEMERAL')
    OUTPUT_TYPES = ('STRING', 'JSON', 'XML', 'HTML')

    def __init__(self, system=None, system_version=None, instance_name=None, command=None,
                 id=None, parent=None, children=None, parameters=None, comment=None, output=None,
                 output_type=None, status=None, command_type=None, created_at=None,
                 error_class=None, metadata=None, updated_at=None):

        self.system = system
        self.system_version = system_version
        self.instance_name = instance_name
        self.command = command
        self.id = id
        self.parent = parent
        self.children = children
        self.parameters = parameters
        self.comment = comment
        self.output = output
        self.output_type = output_type
        self._status = status
        self.command_type = command_type
        self.created_at = created_at
        self.updated_at = updated_at
        self.error_class = error_class
        self.metadata = metadata or {}

    def __str__(self):
        return self.command

    def __repr__(self):
        return ('<Request: command=%s, status=%s, '
                'system=%s, system_version=%s, instance_name=%s>' %
                (self.command, self.status, self.system, self.system_version, self.instance_name))

    @property
    def status(self):
        return self._status

    @property
    def is_ephemeral(self):
        return self.command_type and self.command_type.upper() == 'EPHEMERAL'

    @status.setter
    def status(self, value):
        if self._status in self.COMPLETED_STATUSES:
            raise RequestStatusTransitionError("Status for a request cannot be updated once "
                                               "it has been completed. Current status: {0} "
                                               "Requested status: {1}".format(self.status, value))

        elif (self._status == 'IN_PROGRESS' and
              value not in self.COMPLETED_STATUSES + ('IN_PROGRESS', )):
            raise RequestStatusTransitionError("A request cannot go from IN_PROGRESS to "
                                               "a non-completed status. Completed statuses are "
                                               "{0}. You requested: {1}"
                                               .format(self.COMPLETED_STATUSES, value))
        self._status = value


class System(object):

    schema = 'SystemSchema'

    def __init__(self, name=None, description=None, version=None, id=None, max_instances=None,
                 instances=None, commands=None, icon_name=None, display_name=None, metadata=None):

        self.name = name
        self.description = description
        self.version = version
        self.id = id
        self.max_instances = max_instances
        self.instances = instances or []
        self.commands = commands or []
        self.icon_name = icon_name
        self.display_name = display_name
        self.metadata = metadata or {}

    def __str__(self):
        return '%s-%s' % (self.name, self.version)

    def __repr__(self):
        return '<System: name=%s, version=%s>' % (self.name, self.version)

    @property
    def instance_names(self):
        return [i.name for i in self.instances]

    def has_instance(self, name):
        """Determine if an instance currently exists in the system

        :param name: The name of the instance to search
        :return: True if an instance with the given name exists for this system, False otherwise.
        """
        return True if self.get_instance(name) else False

    def get_instance(self, name):
        """Get an instance that currently exists in the system

        :param name: The name of the instance to search
        :return: The instance with the given name exists for this system, None otherwise
        """
        for instance in self.instances:
            if instance.name == name:
                return instance
        return None

    def get_command_by_name(self, command_name):
        """Retrieve a particular command from the system

        :param command_name: Name of the command to retrieve
        :return: The command object. None if the given command name does not exist in this system.
        """
        for command in self.commands:
            if command.name == command_name:
                return command

        return None

    def has_different_commands(self, commands):
        """Check if a set of commands is different than the current commands

        :param commands: The set commands to compare against the current set
        :return: True if the sets are different, False if the sets are the same
        """
        if len(commands) != len(self.commands):
            return True

        for command in commands:
            if command.name not in [c.name for c in self.commands]:
                return True

            current_command = self.get_command_by_name(command.name)

            if current_command.has_different_parameters(command.parameters):
                return True

        return False


class PatchOperation(object):

    schema = 'PatchSchema'

    def __init__(self, operation=None, path=None, value=None):
        self.operation = operation
        self.path = path
        self.value = value

    def __str__(self):
        return '%s, %s, %s' % (self.operation, self.path, self.value)

    def __repr__(self):
        return '<Patch: operation=%s, path=%s, value=%s>' % (self.operation, self.path, self.value)


class LoggingConfig(object):

    schema = 'LoggingConfigSchema'

    LEVELS = ("DEBUG", "INFO", "WARN", "ERROR")
    SUPPORTED_HANDLERS = ("stdout", "file", "logstash")

    DEFAULT_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    DEFAULT_HANDLER = {
        "class": "logging.StreamHandler",
        "stream": "ext::/sys.stdout",
        "formatter": "default"
    }

    def __init__(self, level=None, handlers=None, formatters=None, loggers=None):
        self.level = level
        self.handlers = handlers
        self.formatters = formatters
        self._loggers = loggers or {}

    @property
    def handler_names(self):
        if self.handlers:
            return self.handlers.keys()
        else:
            return None

    @property
    def formatter_names(self):
        if self.formatters:
            return self.formatters.keys()
        else:
            return None

    def get_plugin_log_config(self, **kwargs):
        """Get a specific plugin logging configuration.

        It is possible for different systems to have different logging configurations. This method
        will create the correct plugin logging configuration and return it. If a specific logger
        is not found for a system, then the current logging configuration will be returned.

        :param kwargs: Identifying information for a system (i.e. system_name)
        :return:
        """
        system_name = kwargs.pop("system_name", None)
        specific_logger = self._loggers.get(system_name, {})

        # If there is no specific logger, then we simply return this object
        # otherwise, we need to construct a new LoggingConfig object with
        # the overrides given in the logger.
        if not specific_logger:
            return self

        level = specific_logger.get("level", self.level)
        handlers = self._generate_handlers(specific_logger.get("handlers"))
        formatters = self._generate_formatters(specific_logger.get("formatters", {}))

        return LoggingConfig(level=level, handlers=handlers, formatters=formatters)

    def _generate_handlers(self, specific_handlers):

        # If we are not given an override for handlers, then we will just
        # assume that we want to use all the handlers given in the current
        # configuration.
        if not specific_handlers:
            return self.handlers

        if isinstance(specific_handlers, list):
            handlers = {}
            for handler_name in specific_handlers:
                handlers[handler_name] = self.handlers[handler_name]
        else:
            return specific_handlers

        return handlers

    def _generate_formatters(self, specific_formatters):

        # If we are not given an override for formatters, then we will just
        # assume that we want to use the formatters given in the current
        # configuration
        if not specific_formatters:
            return self.formatters

        # In case no formatter is provided, we always want a default.
        formatters = {'default': {'format': self.DEFAULT_FORMAT}}
        for formatter_name, format_str in six.iteritems(specific_formatters):
            formatters[formatter_name] = {'format': format_str}

        return formatters

    def __str__(self):
        return '%s, %s, %s' % (self.level, self.handler_names, self.formatter_names)

    def __repr__(self):
        return '<LoggingConfig: level=%s, handlers=%s, formatters=%s' % (self.level,
                                                                         self.handler_names,
                                                                         self.formatter_names)


class Event(object):

    schema = 'EventSchema'

    def __init__(self, name=None, payload=None, error=None, metadata=None, timestamp=None):
        self.name = name
        self.payload = payload
        self.error = error
        self.metadata = metadata or {}
        self.timestamp = timestamp

    def __str__(self):
        return '%s: %s, %s' % (self.name, self.payload, self.metadata)

    def __repr__(self):
        return ('<Event: name=%s, error=%s, payload=%s, metadata=%s>' %
                (self.name, self.error, self.payload, self.metadata))


class Queue(object):

    schema = 'QueueSchema'

    def __init__(self, name=None, system=None, version=None, instance=None, system_id=None,
                 display=None, size=None):
        self.name = name
        self.system = system
        self.version = version
        self.instance = instance
        self.system_id = system_id
        self.display = display
        self.size = size

    def __str__(self):
        return '%s: %s' % (self.name, self.size)

    def __repr__(self):
        return '<Queue: name=%s, size=%s>' % (self.name, self.size)



from lark import Lark, Transformer
from lark.common import ParseError


# Lark added some new errors in later versions
try:
    from lark import GrammarError, LexError
except ImportError:
    GrammarError = ParseError
    LexError = ParseError


choices_grammar = """
    func: CNAME [func_args]
    url: ADDRESS [url_args]
    reference: ref

    func_args: "(" [arg_pair ("," arg_pair)*] ")"
    url_args: "?" arg_pair ("&" arg_pair)*

    arg_pair: CNAME "=" ref
    ?ref: "${" CNAME "}"

    ADDRESS: /^http[^\?]*/

    %import common.CNAME
    %import common.WS
    %ignore WS
"""

parsers = {
    'func': Lark(choices_grammar, start='func'),
    'url': Lark(choices_grammar, start='url'),
    'reference': Lark(choices_grammar, start='reference')
}


class FunctionTransformer(Transformer):

    @staticmethod
    def func(s):
        return {
            'name': str(s[0]),
            'args': s[1] if len(s) > 1 else []
        }

    @staticmethod
    def url(s):
        return {
            'address': str(s[0]),
            'args': s[1] if len(s) > 1 else []
        }

    @staticmethod
    def reference(s):
        return str(s[0])

    @staticmethod
    def arg_pair(s):
        return str(s[0]), str(s[1])

    func_args = list
    url_args = list


def parse(input_string, parse_as=None):
    """Attempt to parse a string into a choices dictionary.

    :param input_string: The string to parse
    :param parse_as: String specifying how to parse `input_string`. Valid values are 'func' or
        'url'. Will try all valid values if None.
    :return: A dictionary containing the results of the parse
    :raise lark.common.ParseError: The parser was not able to find a valid parsing of `input_string`
    """
    def _parse(_input_string, _parser):
        try:
            return FunctionTransformer().transform(_parser.parse(_input_string))
        except (GrammarError, LexError, ParseError) as e:
            raise ParseError(e)

    if parse_as is not None:
        return _parse(input_string, parsers[parse_as])
    else:
        for parser in parsers.values():
            try:
                return _parse(input_string, parser)
            except ParseError:
                continue

    raise ParseError('Unable to successfully parse input "%s"' % input_string)



import functools
import inspect
import json
import os

import requests
import six
import types
import wrapt
from lark.common import ParseError

from brewtils.choices import parse
from brewtils.errors import PluginParamError
from brewtils.models import Command, Parameter, Choices

__all__ = ['system', 'parameter', 'command', 'command_registrar', 'plugin_param', 'register']


# The wrapt module has a cool feature where you can disable wrapping a decorated function,
# instead just using the original function. This is pretty much exactly what we want - we
# aren't using decorators for their 'real' purpose of wrapping a function, we just want to add
# some metadata to the function object. So we'll disable the wrapping normally, but we need to
# test that enabling the wrapping would work.
_wrap_functions = False


def system(cls):
    """Class decorator that marks a class as a beer-garden System

    Creates a ``_commands`` property on the class that holds all registered commands.

    :param cls: The class to decorated
    :return: The decorated class
    """
    commands = []
    for method_name in dir(cls):
        method = getattr(cls, method_name)
        method_command = getattr(method, '_command', None)
        if method_command:
            commands.append(method_command)

    cls._commands = commands

    return cls


def command(_wrapped=None, command_type='ACTION', output_type='STRING', schema=None, form=None,
            template=None, icon_name=None, description=None):
    """Decorator that marks a function as a beer-garden command

    For example:

    .. code-block:: python

        @command(output_type='JSON')
        def echo_json(self, message):
            return message

    :param _wrapped: The function to decorate. This is handled as a positional argument and
        shouldn't be explicitly set.
    :param command_type: The command type. Valid options are Command.COMMAND_TYPES.
    :param output_type: The output type. Valid options are Command.OUTPUT_TYPES.
    :param schema: A custom schema definition.
    :param form: A custom form definition.
    :param template: A custom template definition.
    :param icon_name: The icon name. Should be either a FontAwesome or a Glyphicon name.
    :param description: The command description. Will override the function's docstring.
    :return: The decorated function.
    """
    if _wrapped is None:
        return functools.partial(command, command_type=command_type, output_type=output_type,
                                 schema=schema, form=form, template=template, icon_name=icon_name,
                                 description=description)

    generated_command = _generate_command_from_function(_wrapped)
    generated_command.command_type = command_type
    generated_command.output_type = output_type
    generated_command.icon_name = icon_name

    if description:
        generated_command.description = description

    resolved_mod = _resolve_display_modifiers(_wrapped, generated_command.name, schema=schema,
                                              form=form, template=template)
    generated_command.schema = resolved_mod['schema']
    generated_command.form = resolved_mod['form']
    generated_command.template = resolved_mod['template']

    func_command = getattr(_wrapped, '_command', None)
    if func_command:
        _update_func_command(func_command, generated_command)
    else:
        _wrapped._command = generated_command

    @wrapt.decorator(enabled=_wrap_functions)
    def wrapper(_double_wrapped, _, _args, _kwargs):
        return _double_wrapped(*_args, **_kwargs)

    return wrapper(_wrapped)


def parameter(_wrapped=None, key=None, type=None, multi=None, display_name=None, optional=None,
              default=None, description=None, choices=None, nullable=None, maximum=None,
              minimum=None, regex=None, is_kwarg=None, model=None, form_input_type=None):
    """Decorator that enables Parameter specifications for a beer-garden Command

    This decorator is intended to be used when more specification is desired for a Parameter.

    For example::

        @parameter(key="message", description="Message to echo", optional=True, type="String",
                   default="Hello, World!")
        def echo(self, message):
            return message

    :param _wrapped: The function to decorate. This is handled as a positional argument and
        shouldn't be explicitly set.
    :param key: String specifying the parameter identifier. Must match an argument name of the
        decorated function.
    :param type: String indicating the type to use for this parameter.
    :param multi: Boolean indicating if this parameter is a multi. See documentation for
        discussion of what this means.
    :param display_name: String that will be displayed as a label in the user interface.
    :param optional: Boolean indicating if this parameter must be specified.
    :param default: The value this parameter will be assigned if not overridden when creating a
        request.
    :param description: An additional string that will be displayed in the user interface.
    :param choices: List or dictionary specifying allowed values. See documentation for more
        information.
    :param nullable: Boolean indicating if this parameter is allowed to be null.
    :param maximum: Integer indicating the maximum value of the parameter.
    :param minimum: Integer indicating the minimum value of the parameter.
    :param regex: String describing a regular expression constraint on the parameter.
    :param is_kwarg: Boolean indicating if this parameter is meant to be part of the decorated
        function's kwargs.
    :param model: Class to be used as a model for this parameter. Must be a Python type object,
        not an instance.
    :param form_input_type: Only used for string fields. Changes the form input field
        (e.g. textarea)
    :return: The decorated function.
    """
    if _wrapped is None:
        return functools.partial(parameter, key=key, type=type, multi=multi,
                                 display_name=display_name, optional=optional, default=default,
                                 description=description, choices=choices, nullable=nullable,
                                 maximum=maximum, minimum=minimum, regex=regex, is_kwarg=is_kwarg,
                                 model=model, form_input_type=form_input_type)

    # First see if this method already has a command object associated. If not, create one.
    cmd = getattr(_wrapped, '_command', None)
    if not cmd:
        cmd = _generate_command_from_function(_wrapped)
        _wrapped._command = cmd

    # Every parameter needs a key, so stop that right here
    if key is None:
        raise PluginParamError("Found a parameter definition without a key for "
                               "command '%s'" % cmd.name)

    # If the command doesn't already have a parameter with this key then the method doesn't have
    # an explicit keyword argument with <key> as the name. That's only OK if this parameter is
    # meant to be part of the **kwargs.
    param = cmd.get_parameter_by_key(key)
    if param is None:
        if is_kwarg:
            param = Parameter(key=key, optional=False)
            cmd.parameters.append(param)
        else:
            raise PluginParamError(("Parameter '%s' was not an explicit keyword argument for "
                                    "command '%s' and was not marked as part of kwargs "
                                    "(should is_kwarg be True?)") % (key, cmd.name))

    # Update parameter definition with the plugin_param arguments
    param.type = param.type if type is None else type
    param.multi = param.multi if multi is None else multi
    param.display_name = param.display_name if display_name is None else display_name
    param.optional = param.optional if optional is None else optional
    param.default = param.default if default is None else default
    param.description = param.description if description is None else description
    param.choices = param.choices if choices is None else choices
    param.nullable = param.nullable if nullable is None else nullable
    param.maximum = param.maximum if maximum is None else maximum
    param.minimum = param.minimum if minimum is None else minimum
    param.regex = param.regex if regex is None else regex
    param.form_input_type = param.form_input_type if form_input_type is None else form_input_type

    param.choices = _format_choices(param.choices)

    # Model is another special case - it requires its own handling
    if model is not None:
        param.type = 'Dictionary'
        param.parameters = _generate_nested_params(model)

        # If the model is not nullable and does not have a default defined we will try
        # to generate a default using
        # the defaults defined on the model parameters
        if not param.nullable and not param.default:
            param.default = {}
            for nested_param in param.parameters:
                if nested_param.default:
                    param.default[nested_param.key] = nested_param.default

    @wrapt.decorator(enabled=_wrap_functions)
    def wrapper(_double_wrapped, _, _args, _kwargs):
        return _double_wrapped(*_args, **_kwargs)

    return wrapper(_wrapped)


def _update_func_command(func_command, generated_command):
    """Updates the current function's command with info, (will not override plugin_params)"""
    func_command.name = generated_command.name
    func_command.description = generated_command.description
    func_command.command_type = generated_command.command_type
    func_command.output_type = generated_command.output_type
    func_command.schema = generated_command.schema
    func_command.form = generated_command.form
    func_command.template = generated_command.template
    func_command.icon_name = generated_command.icon_name


def _generate_command_from_function(func):
    """Generates a Command from a function. Uses first line of pydoc as the description."""
    # Required for Python 2/3 compatibility
    if hasattr(func, "func_name"):
        command_name = func.func_name
    else:
        command_name = func.__name__

    # Required for Python 2/3 compatibility
    if hasattr(func, "func_doc"):
        docstring = func.func_doc
    else:
        docstring = func.__doc__

    return Command(name=command_name, description=docstring.split('\n')[0] if docstring else None,
                   parameters=_generate_params_from_function(func))


def _generate_params_from_function(func):
    """Generate Parameters from function arguments.
    Will set the Parameter key, default value, and optional value."""
    parameters_to_return = []

    code = six.get_function_code(func)
    function_arguments = list(code.co_varnames or [])[:code.co_argcount]
    function_defaults = list(six.get_function_defaults(func) or [])

    while len(function_defaults) != len(function_arguments):
        function_defaults.insert(0, None)

    for index, param_name in enumerate(function_arguments):
        # Skip Self or Class reference
        if index == 0 and isinstance(func, types.FunctionType):
            continue

        default = function_defaults[index]
        optional = False if default is None else True

        parameters_to_return.append(Parameter(key=param_name, default=default, optional=optional))

    return parameters_to_return


def _generate_nested_params(model_class):
    """Generates Nested Parameters from a Model Class"""
    parameters_to_return = []
    for parameter_definition in model_class.parameters:
        key = parameter_definition.key
        parameter_type = parameter_definition.type
        multi = parameter_definition.multi
        display_name = parameter_definition.display_name
        optional = parameter_definition.optional
        default = parameter_definition.default
        description = parameter_definition.description
        nullable = parameter_definition.nullable
        maximum = parameter_definition.maximum
        minimum = parameter_definition.minimum
        regex = parameter_definition.regex

        choices = _format_choices(parameter_definition.choices)

        nested_parameters = []
        if parameter_definition.parameters:
            parameter_type = 'Dictionary'
            for nested_class in parameter_definition.parameters:
                nested_parameters = _generate_nested_params(nested_class)

        parameters_to_return.append(Parameter(key=key, type=parameter_type, multi=multi,
                                              display_name=display_name, optional=optional,
                                              default=default, description=description,
                                              choices=choices, parameters=nested_parameters,
                                              nullable=nullable, maximum=maximum, minimum=minimum,
                                              regex=regex))
    return parameters_to_return


def _resolve_display_modifiers(wrapped, command_name, schema=None, form=None, template=None):

    def _load_from_url(url):
        return json.loads(requests.get(url).text)

    def _load_from_path(path):
        current_dir = os.path.dirname(inspect.getfile(wrapped))
        file_path = os.path.abspath(os.path.join(current_dir, path))

        with open(file_path, 'r') as definition_file:
            return definition_file.read()

    resolved = {}

    for key, value in {'schema': schema, 'form': form, 'template': template}.items():

        if isinstance(value, six.string_types):
            try:
                if value.startswith('http'):
                    resolved[key] = _load_from_url(value)

                elif value.startswith('/') or value.startswith('.'):
                    loaded_value = _load_from_path(value)
                    resolved[key] = loaded_value if key == 'template' else json.loads(loaded_value)

                elif key == 'template':
                    resolved[key] = value

                else:
                    raise PluginParamError("%s specified for command '%s' was not a "
                                           "definition, file path, or URL" %
                                           (key, command_name))
            except Exception as ex:
                raise PluginParamError("Error reading %s definition from '%s' for command "
                                       "'%s': %s" % (key, value, command_name, ex))

        elif value is None or (key in ['schema', 'form'] and isinstance(value, dict)):
            resolved[key] = value

        elif key == 'form' and isinstance(value, list):
            resolved[key] = {'type': 'fieldset', 'items': value}

        else:
            raise PluginParamError("%s specified for command '%s' was not a definition, "
                                   "file path, or URL" % (key, command_name))

    return resolved


def _format_choices(choices):

    def determine_display(display_value):
        if isinstance(display_value, six.string_types):
            return 'typeahead'

        return 'select' if len(display_value) <= 50 else 'typeahead'

    def determine_type(type_value):
        if isinstance(type_value, (list, dict)):
            return 'static'
        elif type_value.startswith('http'):
            return 'url'
        else:
            return 'command'

    if not choices:
        return None

    if not isinstance(choices, (list, six.string_types, dict)):
        raise PluginParamError("Invalid 'choices' provided. Must be a list, dictionary or string.")

    elif isinstance(choices, dict):
        if not choices.get('value'):
            raise PluginParamError("No 'value' provided for choices. You must at least "
                                   "provide valid values.")

        value = choices.get('value')
        display = choices.get('display', determine_display(value))
        choice_type = choices.get('type')
        strict = choices.get('strict', True)

        if choice_type is None:
            choice_type = determine_type(value)
        elif choice_type not in Choices.TYPES:
            raise PluginParamError("Invalid choices type '%s' - Valid type options are %s" %
                                   (choice_type, Choices.TYPES))
        else:
            if (choice_type == 'command' and not isinstance(value, (six.string_types, dict))) \
                    or (choice_type == 'url' and not isinstance(value, six.string_types)) \
                    or (choice_type == 'static' and not isinstance(value, (list, dict))):
                allowed_types = {'command': "('string', 'dictionary')", 'url': "('string')",
                                 'static': "('list', 'dictionary)"}
                raise PluginParamError("Invalid choices value type '%s' - Valid value types for "
                                       "choice type '%s' are %s"
                                       % (type(value), choice_type, allowed_types[choice_type]))

        if display not in Choices.DISPLAYS:
            raise PluginParamError("Invalid choices display '%s' - Valid display options are %s" %
                                   (display, Choices.DISPLAYS))
    else:
        value = choices
        display = determine_display(value)
        choice_type = determine_type(value)
        strict = True

    # Now parse out type-specific aspects
    unparsed_value = ''
    try:
        if choice_type == 'command':
            if isinstance(value, six.string_types):
                unparsed_value = value
            else:
                unparsed_value = value['command']

            details = parse(unparsed_value, parse_as='func')
        elif choice_type == 'url':
            unparsed_value = value
            details = parse(unparsed_value, parse_as='url')
        else:
            if isinstance(value, dict):
                unparsed_value = choices.get('key_reference')
                if unparsed_value is None:
                    raise PluginParamError('Specifying a static choices dictionary requires a '
                                           '"key_reference" field with a reference to another '
                                           'parameter ("key_reference": "${param_key}")')

                details = {'key_reference': parse(unparsed_value, parse_as='reference')}
            else:
                details = {}
    except ParseError:
        raise PluginParamError("Invalid choices definition - Unable to parse '%s'" % unparsed_value)

    return Choices(type=choice_type, display=display, value=value, strict=strict, details=details)


# Alias the old names for compatibility
command_registrar = system
plugin_param = parameter
register = command



"""Brewtils Logging Utilities

This module is for setting up your plugins logging correctly.

Example:
    In order to use this, you should simply call ``setup_logger`` in the same file where you
    initialize your plugin sometime before you initialize your Plugin object.

    .. code-block:: python

        host = 'localhost'
        port = 2337
        ssl_enabled = False
        system_name = 'my_system'

        setup_logger(bg_host=host, bg_port=port, system_name=system_name, ssl_enabled=ssl_enabled)
        plugin = Plugin(my_client, bg_host=host, bg_port=port, ssl_enabled=ssl_enabled,
                        name=system_name, version="0.0.1")
        plugin.run()
"""

import logging.config
import copy
import brewtils

# Loggers to always use. These are things that generally,
# people do not want to see and/or are too verbose.
DEFAULT_LOGGERS = {
    "pika": {
        "level": "ERROR"
    },
    "requests.packages.urllib3.connectionpool": {
        "level": "WARN"
    }
}

# A simple default format/formatter. Generally speaking, the API should return formatters,
# but since users can configure their logging, it's better if the formatter has a logical backup.
DEFAULT_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
DEFAULT_FORMATTERS = {
    "default": {
        "format": DEFAULT_FORMAT
    }
}

# A simple default handler. Generally speaking, the API should return handlers, but since
# users can configure their logging, it's better if the handler has a logical backup.
DEFAULT_HANDLERS = {
    "default": {
        "class": "logging.StreamHandler",
        "formatter": "default",
        "stream": "ext://sys.stdout"
    }
}

# The template that plugins will use to log
DEFAULT_PLUGIN_LOGGING_TEMPLATE = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {},
    "handlers": {},
    "loggers": DEFAULT_LOGGERS,
}

# If no logging was configured, this will be used as the logging configuration
DEFAULT_LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": DEFAULT_FORMATTERS,
    "handlers": DEFAULT_HANDLERS,
    "loggers": DEFAULT_LOGGERS,
    "root": {
        "level": "INFO",
        "handlers": ["default"]
    }
}


def setup_logger(bg_host, bg_port, system_name, ca_cert=None, client_cert=None, ssl_enabled=None):
    """Configures python logging module to use logging specified in beer-garden API.

    This method will overwrite your current logging configuration, so only call it if you want
    beer-garden's logging configuration.

    :param str bg_host: Hostname of a beer-garden
    :param int bg_port: Port beer-garden is listening on
    :param str system_name: The system
    :param ca_cert: Certificate that issued the server certificate used by the beer-garden server
    :param client_cert: Certificate used by the server making the connection to beer-garden
    :param bool ssl_enabled: Whether to use SSL for beer-garden communication
    :return:
    """
    config = get_python_logging_config(bg_host=bg_host, bg_port=bg_port, system_name=system_name,
                                       ca_cert=ca_cert, client_cert=client_cert,
                                       ssl_enabled=ssl_enabled)
    logging.config.dictConfig(config)


def get_python_logging_config(bg_host, bg_port, system_name,
                              ca_cert=None, client_cert=None, ssl_enabled=None):
    """Returns a dictionary for the python logging configuration

    :param str bg_host: Hostname of a beer-garden
    :param int bg_port: Port beer-garden is listening on
    :param str system_name: The system
    :param ca_cert: Certificate that issued the server certificate used by the beer-garden server
    :param client_cert: Certificate used by the server making the connection to beer-garden
    :param bool ssl_enabled: Whether to use SSL for beer-garden communication
    :return: Python logging configuration
    """
    client = brewtils.get_easy_client(host=bg_host, port=bg_port, ssl_enabled=ssl_enabled,
                                      ca_cert=ca_cert, client_cert=client_cert)

    logging_config = client.get_logging_config(system_name=system_name)
    return convert_logging_config(logging_config)


def convert_logging_config(logging_config):
    """Converts a LoggingConfig object into a python logging configuration

    The python logging configuration that is returned can be passed to `logging.config.dictConfig`

    :param logging_config:
    :return: Python logging configuration
    """
    config_to_return = copy.deepcopy(DEFAULT_PLUGIN_LOGGING_TEMPLATE)

    if logging_config.handlers:
        handlers = logging_config.handlers
    else:
        handlers = copy.deepcopy(DEFAULT_HANDLERS)
    config_to_return['handlers'] = handlers

    if logging_config.formatters:
        formatters = logging_config.formatters
    else:
        formatters = copy.deepcopy(DEFAULT_FORMATTERS)
    config_to_return['formatters'] = formatters

    config_to_return['root'] = {
        "level": logging_config.level,
        "handlers": list(config_to_return['handlers'].keys())
    }
    return config_to_return




def normalize_url_prefix(url_prefix):

    # Regex to find everything between the / / in the url
    # Should cover all cases
    # url_prefix -------- base_url
    # None                http://localhost:2337/
    # ''                  http://localhost:2337/
    # '/'                 http://localhost:2337/
    # 'example'           http://localhost:2337/example/
    # '/example'          http://localhost:2337/example/
    # 'example/'          http://localhost:2337/example/
    # '/example/'         http://localhost:2337/example/

    if url_prefix in (None, '/', ''):
        return '/'

    new_url_prefix = ""

    # Make string begin with /
    if not url_prefix.startswith("/"):
        new_url_prefix += '/'

    new_url_prefix += url_prefix

    # Make string end with /
    if not url_prefix.endswith("/"):
        new_url_prefix += '/'

    return new_url_prefix



import calendar
import datetime

from marshmallow import Schema, post_dump, post_load, pre_load, fields
from marshmallow.utils import UTC


class DateTime(fields.DateTime):
    """Class that adds methods for (de)serializing DateTime fields as an epoch"""

    def __init__(self, format='epoch', **kwargs):
        self.DATEFORMAT_SERIALIZATION_FUNCS['epoch'] = self.to_epoch
        self.DATEFORMAT_DESERIALIZATION_FUNCS['epoch'] = self.from_epoch
        super(DateTime, self).__init__(format=format, **kwargs)

    @staticmethod
    def to_epoch(dt, localtime=False):
        if localtime and dt.tzinfo is not None:
            localized = dt
        else:
            if dt.tzinfo is None:
                localized = UTC.localize(dt)
            else:
                localized = dt.astimezone(UTC)
        return (calendar.timegm(localized.timetuple()) * 1000) + int(localized.microsecond / 1000)

    @staticmethod
    def from_epoch(epoch):
        # utcfromtimestamp will correctly parse milliseconds in Python 3,
        # but in Python 2 we need to help it
        seconds, millis = divmod(epoch, 1000)
        return datetime.datetime.utcfromtimestamp(seconds).replace(microsecond=millis*1000)


class BaseSchema(Schema):

    def __init__(self, strict=True, **kwargs):
        super(BaseSchema, self).__init__(strict=strict, **kwargs)

    @post_load
    def make_object(self, data):
        try:
            model_class = self.context['models'][self.__class__.__name__]
        except KeyError:
            return data

        return model_class(**data)

    @classmethod
    def get_attribute_names(cls):
        return [key for key, value in cls._declared_fields.items()
                if isinstance(value, fields.FieldABC)]


class ChoicesSchema(BaseSchema):

    type = fields.Str()
    display = fields.Str()
    value = fields.Raw(many=True)
    strict = fields.Bool(default=False)
    details = fields.Dict()


class ParameterSchema(BaseSchema):

    key = fields.Str()
    type = fields.Str(allow_none=True)
    multi = fields.Bool(allow_none=True)
    display_name = fields.Str(allow_none=True)
    optional = fields.Bool(allow_none=True)
    default = fields.Raw(allow_none=True)
    description = fields.Str(allow_none=True)
    choices = fields.Nested('ChoicesSchema', allow_none=True, many=False)
    parameters = fields.Nested('self', many=True, allow_none=True)
    nullable = fields.Bool(allow_none=True)
    maximum = fields.Int(allow_none=True)
    minimum = fields.Int(allow_none=True)
    regex = fields.Str(allow_none=True)
    form_input_type = fields.Str(allow_none=True)


class CommandSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    name = fields.Str()
    description = fields.Str(allow_none=True)
    parameters = fields.Nested('ParameterSchema', many=True)
    command_type = fields.Str(allow_none=True)
    output_type = fields.Str(allow_none=True)
    schema = fields.Dict(allow_none=True)
    form = fields.Dict(allow_none=True)
    template = fields.Str(allow_none=True)
    icon_name = fields.Str(allow_none=True)
    system = fields.Nested('SystemSchema', only=('id', ), allow_none=True)


class InstanceSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    name = fields.Str()
    description = fields.Str(allow_none=True)
    status = fields.Str(allow_none=True)
    status_info = fields.Nested('StatusInfoSchema', allow_none=True)
    queue_type = fields.Str(allow_none=True)
    queue_info = fields.Dict(allow_none=True)
    icon_name = fields.Str(allow_none=True)
    metadata = fields.Dict(allow_none=True)


class SystemSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    name = fields.Str()
    description = fields.Str(allow_none=True)
    version = fields.Str()
    max_instances = fields.Integer(allow_none=True)
    icon_name = fields.Str(allow_none=True)
    instances = fields.Nested('InstanceSchema', many=True, allow_none=True)
    commands = fields.Nested('CommandSchema', many=True)
    display_name = fields.Str(allow_none=True)
    metadata = fields.Dict(allow_none=True)


class RequestSchema(BaseSchema):

    id = fields.Str(allow_none=True)
    system = fields.Str(allow_none=True)
    system_version = fields.Str(allow_none=True)
    instance_name = fields.Str(allow_none=True)
    command = fields.Str(allow_none=True)
    parent = fields.Nested('self', exclude=('children', ), allow_none=True)
    children = fields.Nested('self', exclude=('parent', 'children'), many=True,
                             default=None, allow_none=True)
    parameters = fields.Dict(allow_none=True)
    comment = fields.Str(allow_none=True)
    output = fields.Str(allow_none=True)
    output_type = fields.Str(allow_none=True)
    status = fields.Str(allow_none=True)
    command_type = fields.Str(allow_none=True)
    error_class = fields.Str(allow_none=True)
    created_at = DateTime(allow_none=True, format='epoch', example='1500065932000')
    updated_at = DateTime(allow_none=True, format='epoch', example='1500065932000')
    metadata = fields.Dict(allow_none=True)


class StatusInfoSchema(BaseSchema):

    heartbeat = DateTime(allow_none=True, format='epoch', example='1500065932000')


class PatchSchema(BaseSchema):

    operation = fields.Str()
    path = fields.Str(allow_none=True)
    value = fields.Raw(allow_none=True)

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many):
        if isinstance(data, list):
            return data
        elif 'operations' in data:
            return data['operations']
        else:
            return [data]

    @post_dump(pass_many=True)
    def wrap_envelope(self, data, many):
        return {u'operations': data if many else [data]}


class LoggingConfigSchema(BaseSchema):

    level = fields.Str(allow_none=True)
    formatters = fields.Dict(allow_none=True)
    handlers = fields.Dict(allow_none=True)


class EventSchema(BaseSchema):

    name = fields.Str(allow_none=True)
    payload = fields.Dict(allow_none=True)
    error = fields.Bool(allow_none=True)
    metadata = fields.Dict(allow_none=True)
    timestamp = DateTime(allow_none=True, format='epoch', example='1500065932000')


class QueueSchema(BaseSchema):

    name = fields.Str(allow_none=True)
    system = fields.Str(allow_none=True)
    version = fields.Str(allow_none=True)
    instance = fields.Str(allow_none=True)
    system_id = fields.Str(allow_none=True)
    display = fields.Str(allow_none=True)
    size = fields.Integer(allow_none=True)



import re

from setuptools import setup, find_packages


def find_version():
    version_file = "brewtils/_version.py"
    version_line = open(version_file, "rt").read()
    match_object = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_line, re.M)

    if not match_object:
        raise RuntimeError("Unable to find version string in %s" % version_file)

    return match_object.group(1)


setup(
    name='brewtils',
    version=find_version(),
    description='Utilities for building and running beer-garden Systems',
    url=' ',
    author='The beer-garden Team',
    author_email=' ',
    license='MIT',
    packages=find_packages(exclude=['test', 'test.*']),
    package_data={'': ['README.md']},
    install_requires=[
        'marshmallow',
        'pika',
        'requests',
        'six',
        'wrapt',
        'lark-parser',
        'enum34;python_version<"3.4"',
        'futures;python_version<"3.0"'
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)



import logging
import warnings

import urllib3
from requests import Session

from brewtils.rest import normalize_url_prefix


class RestClient(object):
    """Simple Rest Client for communicating to with beer-garden.

    The is the low-level client responsible for making the actual REST calls. Other clients
    (e.g. :py:class:`brewtils.rest.easy_client.EasyClient`) build on this by providing useful
    abstractions.

    :param host: beer-garden REST API hostname.
    :param port: beer-garden REST API port.
    :param ssl_enabled: Flag indicating whether to use HTTPS when communicating with beer-garden.
    :param api_version: The beer-garden REST API version. Will default to the latest version.
    :param logger: The logger to use. If None one will be created.
    :param ca_cert: beer-garden REST API server CA certificate.
    :param client_cert: The client certificate to use when making requests.
    :param url_prefix: beer-garden REST API Url Prefix.
    :param ca_verify: Flag indicating whether to verify server certificate when making a request.
    """

    # The Latest Version Currently released
    LATEST_VERSION = 1

    JSON_HEADERS = {'Content-type': 'application/json', 'Accept': 'text/plain'}

    def __init__(self, host, port, ssl_enabled=False, api_version=None, logger=None, ca_cert=None,
                 client_cert=None, url_prefix=None, ca_verify=True):
        self.logger = logger or logging.getLogger(__name__)

        # Configure the session to use when making requests
        self.session = Session()

        if not ca_verify:
            urllib3.disable_warnings()
            self.session.verify = False
        elif ca_cert:
            self.session.verify = ca_cert

        if client_cert:
            self.session.cert = client_cert

        # Configure the beer-garden URLs
        scheme = 'https' if ssl_enabled else 'http'
        base_url = '%s://%s:%s%s' % (scheme, host, port, normalize_url_prefix(url_prefix))
        self.version_url = base_url + 'version'
        self.config_url = base_url + 'config'

        api_version = api_version or self.LATEST_VERSION
        if api_version == 1:
            self.system_url = base_url + 'api/v1/systems/'
            self.instance_url = base_url + 'api/v1/instances/'
            self.command_url = base_url + 'api/v1/commands/'
            self.request_url = base_url + 'api/v1/requests/'
            self.queue_url = base_url + 'api/v1/queues/'
            self.logging_config_url = base_url + 'api/v1/config/logging/'
            self.event_url = base_url + 'api/vbeta/events/'
        else:
            raise ValueError("Invalid beer-garden API version: %s" % api_version)

    def get_version(self, **kwargs):
        """Perform a GET to the version URL

        :param kwargs: Parameters to be used in the GET request
        :return: The request response
        """
        return self.session.get(self.version_url, params=kwargs)

    def get_logging_config(self, **kwargs):
        """Perform a GET to the logging config URL

        :param kwargs: Parameters to be used in the GET request
        :return: The request response
        """
        return self.session.get(self.logging_config_url, params=kwargs)

    def get_systems(self, **kwargs):
        """Perform a GET on the System collection URL

        :param kwargs: Parameters to be used in the GET request
        :return: The request response
        """
        return self.session.get(self.system_url, params=kwargs)

    def get_system(self, system_id, **kwargs):
        """Performs a GET on the System URL

        :param system_id: ID of system
        :param kwargs: Parameters to be used in the GET request
        :return: Response to the request
        """
        return self.session.get(self.system_url + system_id, params=kwargs)

    def post_systems(self, payload):
        """Performs a POST on the System URL

        :param payload: New request definition
        :return: Response to the request
        """
        return self.session.post(self.system_url, data=payload, headers=self.JSON_HEADERS)

    def patch_system(self, system_id, payload):
        """Performs a PATCH on a System URL

        :param system_id: ID of system
        :param payload: The update specification
        :return: Response
        """
        return self.session.patch(self.system_url + str(system_id),
                                  data=payload, headers=self.JSON_HEADERS)

    def delete_system(self, system_id):
        """Performs a DELETE on a System URL

        :param system_id: The ID of the system to remove
        :return: Response to the request
        """
        return self.session.delete(self.system_url + system_id)

    def patch_instance(self, instance_id, payload):
        """Performs a PATCH on the instance URL

        :param instance_id: ID of instance
        :param payload: The update specification
        :return: Response
        """
        return self.session.patch(self.instance_url + str(instance_id),
                                  data=payload, headers=self.JSON_HEADERS)

    def get_commands(self):
        """Performs a GET on the Commands URL"""
        return self.session.get(self.command_url)

    def get_command(self, command_id):
        """Performs a GET on the Command URL

        :param command_id: ID of command
        :return: Response to the request
        """
        return self.session.get(self.command_url + command_id)

    def get_requests(self, **kwargs):
        """Performs a GET on the Requests URL

        :param kwargs: Parameters to be used in the GET request
        :return: Response to the request
        """
        return self.session.get(self.request_url, params=kwargs)

    def get_request(self, request_id):
        """Performs a GET on the Request URL

        :param request_id: ID of request
        :return: Response to the request
        """
        return self.session.get(self.request_url + request_id)

    def post_requests(self, payload):
        """Performs a POST on the Request URL

        :param payload: New request definition
        :return: Response to the request
        """
        return self.session.post(self.request_url, data=payload, headers=self.JSON_HEADERS)

    def patch_request(self, request_id, payload):
        """Performs a PATCH on the Request URL

        :param request_id: ID of request
        :param payload: New request definition
        :return: Response to the request
        """
        return self.session.patch(self.request_url + str(request_id),
                                  data=payload, headers=self.JSON_HEADERS)

    def post_event(self, payload, publishers=None):
        """Performs a POST on the event URL

        :param payload: New event definition
        :param publishers: Array of publishers to use
        :return: Response to the request
        """
        return self.session.post(self.event_url, data=payload, headers=self.JSON_HEADERS,
                                 params={'publisher': publishers} if publishers else None)

    def get_queues(self):
        """Performs a GET on the Queues URL

        :return: Response to the request
        """
        return self.session.get(self.queue_url)

    def delete_queues(self):
        """Performs a DELETE on the Queues URL

        :return: Response to the request
        """
        return self.session.delete(self.queue_url)

    def delete_queue(self, queue_name):
        """Performs a DELETE on a specific Queue URL

        :return: Response to the request
        """
        return self.session.delete(self.queue_url + queue_name)


class BrewmasterRestClient(RestClient):
    def __init__(self, *args, **kwargs):
        warnings.warn("Call made to 'BrewmasterRestClient'. This name will be removed in version "
                      "3.0, please use 'RestClient' instead.", DeprecationWarning, stacklevel=2)
        super(BrewmasterRestClient, self).__init__(*args, **kwargs)



import logging
import warnings

from brewtils.errors import BrewmasterFetchError, BrewmasterValidationError, BrewmasterSaveError, \
    BrewmasterDeleteError, BrewmasterConnectionError, BGNotFoundError, BGConflictError, \
    BrewmasterRestError
from brewtils.models import Event, PatchOperation
from brewtils.rest.client import RestClient
from brewtils.schema_parser import SchemaParser


class EasyClient(object):
    """Client for communicating with beer-garden.

    This class provides nice wrappers around the functionality provided by a
    :py:class:`brewtils.rest.client.RestClient`

    :param host: beer-garden REST API hostname.
    :param port: beer-garden REST API port.
    :param ssl_enabled: Flag indicating whether to use HTTPS when communicating with beer-garden.
    :param api_version: The beer-garden REST API version. Will default to the latest version.
    :param ca_cert: beer-garden REST API server CA certificate.
    :param client_cert: The client certificate to use when making requests.
    :param parser: The parser to use. If None will default to an instance of BrewmasterSchemaParser.
    :param logger: The logger to use. If None one will be created.
    :param url_prefix: beer-garden REST API URL Prefix.
    :param ca_verify: Flag indicating whether to verify server certificate when making a request.
    """

    def __init__(self, host, port, ssl_enabled=False, api_version=None, ca_cert=None,
                 client_cert=None, parser=None, logger=None, url_prefix=None, ca_verify=True):
        self.logger = logger or logging.getLogger(__name__)
        self.parser = parser or SchemaParser()
        self.client = RestClient(host=host, port=port, ssl_enabled=ssl_enabled,
                                 api_version=api_version, ca_cert=ca_cert, client_cert=client_cert,
                                 url_prefix=url_prefix, ca_verify=ca_verify)

    def get_version(self, **kwargs):
        response = self.client.get_version(**kwargs)
        if response.ok:
            return response
        else:
            self._handle_response_failure(response, default_exc=BrewmasterFetchError)

    def find_unique_system(self, **kwargs):
        """Find a unique system using keyword arguments as search parameters.

        :param kwargs: Search parameters
        :return: One system instance
        """
        if 'id' in kwargs:
            return self._find_system_by_id(kwargs.pop('id'), **kwargs)
        else:
            systems = self.find_systems(**kwargs)

            if not systems:
                return None

            if len(systems) > 1:
                raise BrewmasterFetchError("More than one system found that specifies "
                                           "the given constraints")

            return systems[0]

    def find_systems(self, **kwargs):
        """Find systems using keyword arguments as search parameters.

        :param kwargs: Search parameters
        :return: A list of system instances satisfying the given search parameters
        """
        response = self.client.get_systems(**kwargs)

        if response.ok:
            return self.parser.parse_system(response.json(), many=True)
        else:
            self._handle_response_failure(response, default_exc=BrewmasterFetchError)

    def _find_system_by_id(self, system_id, **kwargs):
        """Finds a system by id, convert JSON to a system object and return it."""
        response = self.client.get_system(system_id, **kwargs)

        if response.ok:
            return self.parser.parse_system(response.json())
        else:
            self._handle_response_failure(response, default_exc=BrewmasterFetchError,
                                          raise_404=False)

    def create_system(self, system):
        """Create a new system by POSTing to a BREWMASTER server.

        :param system: The system to create
        :return: The system creation response
        """
        json_system = self.parser.serialize_system(system)
        response = self.client.post_systems(json_system)

        if response.ok:
            return self.parser.parse_system(response.json())
        else:
            self._handle_response_failure(response, default_exc=BrewmasterSaveError)

    def update_system(self, system_id, new_commands=None, **kwargs):
        """Update a system with a PATCH

        :param system_id: The ID of the system to update
        :param new_commands: The new commands

        :Keyword Arguments:
            * *metadata* (``dict``) The updated metadata for the system
            * *description* (``str``) The updated description for the system
            * *display_name* (``str``) The updated display_name for the system
            * *icon_name* (``str``) The updated icon_name for the system

        :return: The response
        """
        operations = []
        metadata = kwargs.pop("metadata", {})

        if new_commands:
            operations.append(PatchOperation('replace', '/commands',
                                             self.parser.serialize_command(new_commands,
                                                                           to_string=False,
                                                                           many=True)))

        if metadata:
            operations.append(PatchOperation('update', '/metadata', metadata))

        for attr, value in kwargs.items():
            if value is not None:
                operations.append(PatchOperation('replace', '/%s' % attr, value))

        response = self.client.patch_system(system_id, self.parser.serialize_patch(operations,
                                                                                   many=True))

        if response.ok:
            return self.parser.parse_system(response.json())
        else:
            self._handle_response_failure(response, default_exc=BrewmasterSaveError)

    def remove_system(self, **kwargs):
        """Remove a specific system using keyword arguments as search parameters.

        :param kwargs: Search parameters
        :return: The response
        """
        system = self.find_unique_system(**kwargs)

        if system is None:
            raise BrewmasterFetchError("Could not find system matching the given search parameters")

        return self._remove_system_by_id(system.id)

    def _remove_system_by_id(self, system_id):

        if system_id is None:
            raise BrewmasterDeleteError("Cannot delete a system without an id")

        response = self.client.delete_system(system_id)
        if response.ok:
            return True
        else:
            self._handle_response_failure(response, default_exc=BrewmasterDeleteError)

    def initialize_instance(self, instance_id):
        """Start an instance by PATCHing to a BREWMASTER server.

        :param instance_id: The ID of the instance to start
        :return: The start response
        """
        response = self.client.patch_instance(instance_id,
                                              self.parser.serialize_patch(
                                                  PatchOperation('initialize')
                                              ))

        if response.ok:
            return self.parser.parse_instance(response.json())
        else:
            self._handle_response_failure(response, default_exc=BrewmasterSaveError)

    def update_instance_status(self, instance_id, new_status):
        """Update an instance by PATCHing to a BREWMASTER server.

        :param instance_id: The ID of the instance to start
        :param new_status: The updated status
        :return: The start response
        """
        payload = PatchOperation('replace', '/status', new_status)
        response = self.client.patch_instance(instance_id, self.parser.serialize_patch(payload))

        if response.ok:
            return self.parser.parse_instance(response.json())
        else:
            self._handle_response_failure(response, default_exc=BrewmasterSaveError)

    def instance_heartbeat(self, instance_id):
        """Send heartbeat to BREWMASTER for health and status purposes

        :param instance_id: The ID of the instance
        :return: The response
        """
        payload = PatchOperation('heartbeat')
        response = self.client.patch_instance(instance_id, self.parser.serialize_patch(payload))

        if response.ok:
            return True
        else:
            self._handle_response_failure(response, default_exc=BrewmasterSaveError)

    def find_unique_request(self, **kwargs):
        """Find a unique request using keyword arguments as search parameters.

        .. note::
            If 'id' is present in kwargs then all other parameters will be ignored.

        :param kwargs: Search parameters
        :return: One request instance
        """
        if 'id' in kwargs:
            return self._find_request_by_id(kwargs.pop('id'))
        else:
            requests = self.find_requests(**kwargs)

            if not requests:
                return None

            if len(requests) > 1:
                raise BrewmasterFetchError("More than one request found that specifies "
                                           "the given constraints")

            return requests[0]

    def find_requests(self, **kwargs):
        """Find requests using keyword arguments as search parameters.

        :param kwargs: Search parameters
        :return: A list of request instances satisfying the given search parameters
        """
        response = self.client.get_requests(**kwargs)

        if response.ok:
            return self.parser.parse_request(response.json(), many=True)
        else:
            self._handle_response_failure(response, default_exc=BrewmasterFetchError)

    def _find_request_by_id(self, request_id):
        """Finds a request by id, convert JSON to a request object and return it."""
        response = self.client.get_request(request_id)

        if response.ok:
            return self.parser.parse_request(response.json())
        else:
            self._handle_response_failure(response, default_exc=BrewmasterFetchError,
                                          raise_404=False)

    def create_request(self, request):
        """Create a new request.

        :param request: The request to create
        :return: The response
        """
        json_request = self.parser.serialize_request(request)
        response = self.client.post_requests(json_request)

        if response.ok:
            return self.parser.parse_request(response.json())
        else:
            self._handle_response_failure(response, default_exc=BrewmasterSaveError)

    def update_request(self, request_id, status=None, output=None, error_class=None):
        """Set various fields on a request with a PATCH

        :param request_id: The ID of the request to update
        :param status: The new status
        :param output: The new output
        :param error_class: The new error class
        :return: The response
        """
        operations = []

        if status:
            operations.append(PatchOperation('replace', '/status', status))
        if output:
            operations.append(PatchOperation('replace', '/output', output))
        if error_class:
            operations.append(PatchOperation('replace', '/error_class', error_class))

        response = self.client.patch_request(request_id, self.parser.serialize_patch(operations,
                                                                                     many=True))

        if response.ok:
            return self.parser.parse_request(response.json())
        else:
            self._handle_response_failure(response, default_exc=BrewmasterSaveError)

    def get_logging_config(self, system_name):
        """Get the logging configuration for a particular system.

        :param system_name: Name of system
        :return: LoggingConfig object
        """
        response = self.client.get_logging_config(system_name=system_name)
        if response.ok:
            return self.parser.parse_logging_config(response.json())
        else:
            self._handle_response_failure(response, default_exc=BrewmasterConnectionError)

    def publish_event(self, *args, **kwargs):
        """Publish a new event.

        :param args: The Event to create
        :param _publishers: Optional list of specific publishers. If None all publishers will be
            used.
        :param kwargs: If no Event is given in the *args, on will be constructed from the kwargs
        :return: The response
        """
        publishers = kwargs.pop('_publishers', None)
        json_event = self.parser.serialize_event(args[0] if args else Event(**kwargs))

        response = self.client.post_event(json_event, publishers=publishers)

        if response.ok:
            return True
        else:
            self._handle_response_failure(response)

    def get_queues(self):
        """Retrieve all queue information

        :return: The response
        """
        response = self.client.get_queues()

        if response.ok:
            return self.parser.parse_queue(response.json(), many=True)
        else:
            self._handle_response_failure(response)

    def clear_queue(self, queue_name):
        """Cancel and clear all messages from a queue

        :return: The response
        """
        response = self.client.delete_queue(queue_name)

        if response.ok:
            return True
        else:
            self._handle_response_failure(response)

    def clear_all_queues(self):
        """Cancel and clear all messages from all queues

        :return: The response
        """
        response = self.client.delete_queues()

        if response.ok:
            return True
        else:
            self._handle_response_failure(response)

    @staticmethod
    def _handle_response_failure(response, default_exc=BrewmasterRestError, raise_404=True):
        if response.status_code == 404:
            if raise_404:
                raise BGNotFoundError(response.json())
            else:
                return None
        elif response.status_code == 409:
            raise BGConflictError(response.json())
        elif 400 <= response.status_code < 500:
            raise BrewmasterValidationError(response.json())
        elif response.status_code == 503:
            raise BrewmasterConnectionError(response.json())
        else:
            raise default_exc(response.json())


class BrewmasterEasyClient(EasyClient):
    def __init__(self, *args, **kwargs):
        warnings.warn("Call made to 'BrewmasterEasyClient'. This name will be removed in version "
                      "3.0, please use "
                      "'EasyClient' instead.", DeprecationWarning, stacklevel=2)
        super(BrewmasterEasyClient, self).__init__(*args, **kwargs)
